function [z,M,sizeH]=genlattice_cbc_incr_bisect(h,opts)
%
%   [z]=genlattice_cbc_incr_bisect(h,M,opts)
%   h:  frequency index set
%   M:  lattice size
%   opts:
%       opts.Inc [{0} | 1 | rand] 
%           - 0:    search for each component starts from 1
%           - 1:    search for each component z_s starts from z_{s-1}+1,
%                   result is an ordered vector [usefull for equalweighted
%                   hyperbolic crosses or l1balls]
%           - 2:    fixing z(s)=M(s-1)
%           - rand: search randomly each component from 1,...,M-1
%           
%       opts.Display [ {off} | cbc | final ]
%           - off:      no information is shown
%           - cbc:      displays the components at the moment you find it
%           - final:    displays the resulting vector, if exists, and the
%                       computational time
%           - zstart:   lower dimensional vector used as starting point


% first tests:
sh=size(h);
nthreads = 1;
if isfield(opts,'nthreads')
  nthreads = opts.nthreads;
end

if(nargin<2)
    opts.Display='off';
    opts.Inc=0;
else
    if(~(isfield(opts,'Inc') && (max(opts.Inc)<=2 || strcmp(opts.Inc,'rand')==1)))
        opts.Inc=0;
    end
    if(~(isfield(opts,'Display') && strmatch(opts.Display, char('off', 'cbc', 'final'),'exact')>0))
        opts.Display='off';
    end
end

h=double(h);

%M(1)=max(h(:,1))-min(h(:,1))+1;
%sizeH = zeros(1, sh(2));
%sizeH(1) = M(1);
%z(1)=1;
M(1)=max(h(:,1))-min(h(:,1))+1;
z(1)=1;
htmp=unique(h(:,1),'rows');
for j=M(1):-1:size(htmp)
    if(length(unique(mod(htmp,j)))==length(htmp))
        Mtmp=j;
    end
end
M(1)=Mtmp;

ticsearch=tic;
for s=2:size(h,2)
   
    hs=unique(h(:,1:s),'rows');
    sizeH(s) = size(hs,1);
%    P(s)=M(s-1)*(max(h(:,1))-min(h(:,1))+1);
    S(s)=max(h(:,1))-min(h(:,1))+1;
    htmp=unique(h(:,s),'rows');
    for j=S(s):-1:size(htmp)
        if(length(unique(mod(htmp,j)))==length(htmp))
            Stmp=j;
        end
    end
    S(s)=Stmp;
    fprintf('dimension %d, S_%d=%d\n', [s,s,S(s)])
    P(s)=M(s-1)*S(s);

    tmpstart=1;
    switch(opts.Inc)
        case 1
            tmpstart=(z(s-1)+1);
        case 2
            tmpstart=M(s-1);
    end
    for l=tmpstart:M(s-1)
        if(strcmp(opts.Inc,'rand')==1)
            zs=randi(P(s)-1,1);
        else
            zs=l;
        end
        if  length(unique(mod(hs*[z,zs]',P(s))))==size(hs,1)
           z=[z,zs];
           break;
        end
        if(mod(zs,ceil(P(s)/20))==0 && strcmp(opts.Display,'cbc')==1 && strcmp(opts.Inc,'rand')==0)
            fprintf('dimension %d, z_%d>%d\n', [s,s,l])
        end
    end
    if(strcmp(opts.Display,'cbc')==1)
        fprintf('dimension %d, z_%d=%d\n', [s,s,z(s)])
    end
    
    optPmin.Display='xs';
    optPmin.Prime=0;
    optPmin.Bisection_Successors=1000;
    P_in=P(s);
    if nthreads > 1
      optPmin.nthreads = nthreads;
    end
    P_out=lfft_lattice_size_reduce_bisect_only_mex(hs,P_in,z,optPmin);
    fprintf('np start, P_in: %d, P_out: %d, abs: %d, rel: %.3f\n', P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
    P_in=P_out;

    optPmin.Display='iter';
    optPmin.Prime=1;
    optPmin.Bisection_Successors=1000;
%    if P_in > 2^31-1
%      fprintf('Warning: P_in = %d > 2^31-1\n', P_in);
%      P_in = 2^31-1;
%    end
    for r=1:10
      try
        if nthreads > 1
          optPmin.nthreads = nthreads;
        end
        P_out=lfft_lattice_size_reduce_bisect_only_mex(hs,P_in,z,optPmin);
      catch err
        P_out=P_in;
      end
      fprintf('p  r: %d, P_in: %d, P_out: %d, abs: %d, rel: %.3f\n', r, P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
%      if ( ((P_in-P_out)/P_in < 0.01) && (P_in-P_out < 100000))
      if P_in==P_out
        break;
      end
      P_in = P_out;
    end
    Pmin(s)=P_out;

    optM.Display='iter';
    optM.Prime=0;
    optM.Bisection_Successors=1000;
    P_in=Pmin(s);
    for r=1:10
      if nthreads > 1
        optM.nthreads = nthreads;
      end
      P_out=lfft_lattice_size_reduce_bisect_only_mex(hs,P_in,z,optM);
      fprintf('np r: %d, P_in: %d, P_out: %d, abs: %d, rel: %.3f\n', r, P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
%      if ( ((P_in-P_out)/P_in < 0.01) && (P_in-P_out < 100000))
      if P_in==P_out
        break;
      end
      P_in = P_out;
    end
    M(s)=P_out;
%     fprintf('%d\n',M(s));

% %    M(s)=lfft_lattice_size_reduce_bisect(hs,M(s),z,optM);
% %    fprintf('%d\n',M(s));
if s < size(h,2)
     optM.Prime=1;
     optM.Display='verbose';
%     optM.Mmin=M(s-1);
     optM.Mmin=max(ceil(0.85*M(s)), M(s)-200000);
     optM.Mmin=max(optM.Mmin, sizeH(s));
     try
       if nthreads > 1
         optM.nthreads = nthreads;
       end
       M(s)=lfft_lattice_size_reduce_mex(hs,M(s),z,optM);
     catch err
     end
%     M_start = max(M(s-1),sizeH(s));
%     if exist(filename, 'file') > 0
%       load(filename);
%       ii=find(sizeH_matrix(:,s) < sizeH(s));
%       if length(ii) > 0
%         ii2=sortrows([sizeH_matrix(ii,s), ii]);
%         ii3=ii2(end,2);
%         factor = M_matrix(ii3,s)/sizeH_matrix(ii3,s);
%         M_start2 = factor*sizeH(s);
%         M_start = max(M_start, M_start2);
%       end
%     end
%     try
%       M(s)=lattice_size_increase_mex(hs,M_start,z,optM,M(s));
%     catch err
%     end
    optM.Prime=0;
    optM.Display='xs';
    optM.Mmin=max(ceil(0.85*M(s)), M(s)-100000);
    optM.Mmin=max(optM.Mmin, sizeH(s));
%     if s == size(h,2)
%       optM.Mmin=max( [size(hs,1), ceil(0.85*M(s)), ceil(M(s)-200000)] );
%     else
%       optM.Mmin=max( size(hs,1), ceil(0.85*M(s)) );
%     end
    if nthreads > 1
      optM.nthreads = nthreads;
    end
    M(s)=lfft_lattice_size_reduce_mex(hs,M(s),z,optM);
end
    z=mod(z,M(s));
    if(strcmp(opts.Display,'cbc')==1)
        fprintf('dimension %d: P=%d, Pmin=%d, M=%d\n', [s,P(s),Pmin(s),M(s)])
        fprintf('dimension %d, z_%d mod M_s=%d\n', [s,s,z(s)])
    end

    if length(unique(mod(hs*z',M(s))))~=size(hs,1)
      if length(unique(mod(hs*z',Pmin(s))))~=size(hs,1)
        error('Internal error: M(s) and Pmin(s) lattice is not reconstructing lattice!');
      end
      error('Internal error: M(s) lattice is not reconstructing lattice!');
    end
end
t1=toc(ticsearch);

if(strcmp(opts.Display,'final')==1 || strcmp(opts.Display,'cbc')==1)
    strtoshow=sprintf('generating vector z=[%d',z(1));
    for s=2:length(z)
        strtoshow=[strtoshow,',%d'];
    end
    strtoshow=sprintf([strtoshow,']\n'],z(2:length(z)));
    strtoshow=[strtoshow,sprintf('cbc search time: %d',t1)];
    disp(strtoshow);
end

end
